@@ -40,6 +40,7 @@ test_unit(void)
   double x, y, nan, inf;
   IPAddr ip, ip2, ip3;
   IPSockAddr ip_saddr;
+  Integer64 integer64;
   Timespec tspec;
   Float f;
   int i, j, c;
@@ -470,6 +471,15 @@ test_unit(void)
   s = UTI_IPSockAddrToString(&ip_saddr);
   TEST_CHECK(strcmp(s, "1.2.3.4:12345") == 0);
 
+  ip = ip_saddr.ip_addr;
+  s = UTI_IPSubnetToString(&ip, 10);
+  TEST_CHECK(strcmp(s, "1.2.3.4/10") == 0);
+  s = UTI_IPSubnetToString(&ip, 32);
+  TEST_CHECK(strcmp(s, "1.2.3.4") == 0);
+  ip.family = IPADDR_UNSPEC;
+  s = UTI_IPSubnetToString(&ip, 0);
+  TEST_CHECK(strcmp(s, "any address") == 0);
+
   s = UTI_TimeToLogForm(2000000000);
   TEST_CHECK(strcmp(s, "2033-05-18 03:33:20") == 0);
 
@@ -556,6 +566,10 @@ test_unit(void)
   UTI_TimespecNetworkToHost(&tspec, &ts2);
   TEST_CHECK(!UTI_CompareTimespecs(&ts, &ts2));
 
+  integer64 = UTI_Integer64HostToNetwork(0x1234567890ABCDEFULL);
+  TEST_CHECK(memcmp(&integer64, "\x12\x34\x56\x78\x90\xab\xcd\xef", 8) == 0);
+  TEST_CHECK(UTI_Integer64NetworkToHost(integer64) == 0x1234567890ABCDEFULL);
+
   TEST_CHECK(UTI_CmacNameToAlgorithm("AES128") == CMC_AES128);
   TEST_CHECK(UTI_CmacNameToAlgorithm("AES256") == CMC_AES256);
   TEST_CHECK(UTI_CmacNameToAlgorithm("NOSUCHCMAC") == CMC_INVALID);
@@ -741,4 +755,6 @@ test_unit(void)
   TEST_CHECK(words[1] == buf + 3);
   TEST_CHECK(strcmp(words[0], "a") == 0);
   TEST_CHECK(strcmp(words[1], "b") == 0);
+
+  HSH_Finalise();
 }
